package sdds;

import java.util.Iterator;

/**
 * <p>A simple dynamic {@link HashTable}.</p>
 * <p>In general, the time for each operation is as follows:</p>
 * <ul>
 * <li>add an element: O(1) amortized time</li>
 * <li>search for an element: O(1) amortized time</li>
 * <li>remove an element: O(1) amortized time</li>
 * </ul>
 * 
 * @author Stephen G. Ware
 * @param <K> the type of object that will act as keys for the hash table
 * @param <V> the type of object that is being stored in the hash table
 */
public class DHashTable<K, V> extends HashTable<K, V> {
	
	/**
	 * <p>Construct a dynamic hash table with a given number of mappings and given set of buckets.</p>
	 * 
	 * @param size the number of mappings in the hash table
	 * @param table the buckets
	 */
	protected DHashTable(int size, Node<K, V>[] table){
		super(size, table);
	}

	/**
	 * <p>Construct an empty dynamic hash table.</p>
	 */
	public DHashTable(){
		super();
	}
	
	/**
	 * <p>Construct a dynamic hash table which initially contains mappings for a given set of keys and values.</p>
	 * <p>Mappings are added as if using {@link #add}.</p>
	 * 
	 * @param keys the objects which can be used to retrieve the values
	 * @param values the objects to be retrieved using the keys
	 * @throws ArrayIndexOutOfBoundsException if the array of keys is not the same length as the array of values
	 */
	public DHashTable(K[] keys, V[] values){
		super(keys, values);
	}

	/**
	 * <p>Return the load factor of the dynamic hash table.</p>
	 * <p>The load factor is defined as the number of mapping divided by the number of buckets.</p>
	 * 
	 * @return the load factor
	 */
	protected synchronized final float getLoadFactor(){
		return super.getLoadFactor();
	}
	
	/**
	 * <p>Return the number of mappings in the dynamic hash table.</p>
	 * 
	 * @return the number of mappings
	 */
	public synchronized int size(){
		return size;
	}
	
	/**
	 * <p>Add a key/value mapping to the dynamic hash table.  If a mapping exists whose key is <code>==</code> to this key, it is replaced.</p>
	 * 
	 * @param key the object which can be used to retrieve the value
	 * @param value the object to be retrieved using the key
	 * @return this hash table
	 */
	public synchronized DHashTable<K, V> add(K key, V value){
		if(getLoadFactor() > DEFAULT_LOAD_FACTOR_THRESHOLD) table = rehash();
		int bucket = getBucket(key, table);
		Node<K, V> current = table[bucket];
		while(current != null){
			if(current.key == key){
				current.key = key;
				current.value = value;
				return this;
			}
			current = current.next;
		}
		table[bucket] = new Node<K, V>(key, value, table[bucket]);
		size++;
		return this;
	}
	
	/**
	 * <p>Add multiple key/value mappings to the dynamic hash table.  If mappings exist whose keys are <code>==</code> to the given keys, they are replaced.</p>
	 * 
	 * @param keys the objects which can be used to retrieve the values
	 * @param values the objects to be retrieved using the keys
	 * @return this hash table
	 * @throws ArrayIndexOutOfBoundsException if the array of keys is not the same length as the array of values
	 */
	public synchronized DHashTable<K, V> add(K[] keys, V[] values){
		if(keys.length != values.length)
			throw new ArrayIndexOutOfBoundsException("The array of keys must be the same length as the array of values.");
		for(int i=0; i<keys.length; i++)
			add(keys[i], values[i]);
		return this;
	}
	
	/**
	 * <p>Add a key/value mapping to the dynamic hash table.  If a mapping exists whose key {@link Object#equals} this key, it is replaced.</p>
	 * 
	 * @param key the object which can be used to retrieve the value
	 * @param value the object to be retrieved using the key
	 * @return this hash table
	 */
	public synchronized DHashTable<K, V> addEquals(K key, V value){
		if(getLoadFactor() > DEFAULT_LOAD_FACTOR_THRESHOLD) table = rehash();
		int bucket = getBucket(key, table);
		Node<K, V> current = table[bucket];
		while(current != null){
			if(current.key.equals(key)){
				current.key = key;
				current.value = value;
				return this;
			}
			current = current.next;
		}
		table[bucket] = new Node<K, V>(key, value, table[bucket]);
		size++;
		return this;
	}
	
	/**
	 * <p>Add multiple key/value mappings to the dynamic hash table.  If mappings exist whose keys are {@link Object#equals} to the given keys, they are replaced.</p>
	 * 
	 * @param keys the objects which can be used to retrieve the values
	 * @param values the objects to be retrieved using the keys
	 * @return this hash table
	 * @throws ArrayIndexOutOfBoundsException if the array of keys is not the same length as the array of values
	 */
	public synchronized DHashTable<K, V> addEquals(K[] keys, V[] values){
		if(keys.length != values.length)
			throw new ArrayIndexOutOfBoundsException("The array of keys must be the same length as the array of values.");
		for(int i=0; i<keys.length; i++)
			addEquals(keys[i], values[i]);
		return this;
	}

	/**
	 * <p>Return the value mapped to a given key.  Keys are searched using <code>==</code>.</p>
	 * 
	 * @param key the object that was mapped to the desired value
	 * @return the value
	 */
	public synchronized V get(K key){
		return super.get(key);
	}

	/**
	 * <p>Return the value mapped to a given key.  Keys are searched using {@link Object#equals}.</p>
	 * 
	 * @param key the object that was mapped to the desired value
	 * @return the value
	 */
	public synchronized V getEquals(K key){
		return super.getEquals(key);
	}
	
	/**
	 * <p>Remove a key/value mapping from the dynamic hash table.  Keys are searched using <code>==</code>.</p>
	 * 
	 * @param key the key of the key/value mapping to be removed
	 * @return this hash table
	 */
	public synchronized DHashTable<K, V> remove(K key){
		int bucket = getBucket(key, table);
		Node<K, V> current = table[bucket];
		if(current == null) return this;
		else if(current.key == key){
			table[bucket] = current.next;
			size--;
		}
		else{
			while(current.next != null){
				if(current.next.key == key){
					current.next = current.next.next;
					size--;
					return this;
				}
				current = current.next;
			}
		}
		return this;
	}

	/**
	 * <p>Remove multiple key/value mappings from the dynamic hash table.  Keys are searched using <code>==</code>.</p>
	 * 
	 * @param keys the keys of the key/value mappings to be removed
	 * @return this hash table
	 */
	public synchronized DHashTable<K, V> remove(K[] keys){
		for(int i=0; i<keys.length; i++)
			remove(keys[i]);
		return this;
	}

	/**
	 * <p>Remove a key/value mapping from the dynamic hash table.  Keys are searched using {@link Object#equals}.</p>
	 * 
	 * @param key the key of the key/value mapping to be removed
	 * @return this hash table
	 */
	public synchronized DHashTable<K, V> removeEquals(K key){
		int bucket = getBucket(key, table);
		Node<K, V> current = table[bucket];
		if(current == null) return this;
		else if(current.key.equals(key)){
			table[bucket] = current.next;
			size--;
		}
		else{
			while(current.next != null){
				if(current.next.key.equals(key)){
					current.next = current.next.next;
					size--;
					return this;
				}
				current = current.next;
			}
		}
		return this;
	}
	
	/**
	 * <p>Remove multiple key/value mappings from the dynamic hash table.  Keys are searched using {@link Object#equals}.</p>
	 * 
	 * @param keys the keys of the key/value mappings to be removed
	 * @return this hash table
	 */
	public synchronized DHashTable<K ,V> removeEquals(K[] keys){
		for(int i=0; i<keys.length; i++)
			removeEquals(keys[i]);
		return this;
	}
	
	/**
	 * <p>Return an {@link java.util.Iterator} that will yield all the keys in the dynamic hash table.</p>
	 * 
	 * @return an iterator of keys
	 */
	public synchronized Iterator<K> keys(){
		return super.keys();
	}
	
	/**
	 * <p>Return an {@link java.util.Iterator} that will yield all the values in the dynamic hash table.</p>
	 * 
	 * @return an iterator of values
	 */
	public synchronized Iterator<V> values(){
		return super.values();
	}
	
	/**
	 * <p>Check if this hash table is the same as another.</p>
	 * <p>Two hash tables are considered the same if they contain the same key/value mappings.  Keys and values are compared using {@link Object#equals}.</p>
	 * <p>The class of the hash table is not taken into account.  In other words, a {@link DHashTable} can be equal to an {@link SHashTable} if they contain the same key/value mappings.</p>
	 * 
	 * @param other another hash table
	 * @return <tt>true</tt> is the hash tables are the same, <tt>false</tt> if the hash tables are different or if <tt>other</tt> is not a hash table
	 */
	public synchronized boolean equals(Object other){
		return super.equals(other);
	}
	
	/**
	 * <p>Return a deep copy of the dynamic hash table.</p>
	 * <p>The clone hash table can be modified without affecting the original.</p>
	 */
	public synchronized DHashTable<K, V> clone(){
		return new DHashTable<K, V>(size, cloneTable());
	}
	
	/**
	 * <p>Return a string representation of this hash table and all its elements.</p>
	 */
	public synchronized String toString(){
		return super.toString();
	}
}
