package sdds;

import java.util.Iterator;

/**
 * <p>A dynamic singly-linked {@link List}.</p>
 * 
 * @author Stephen G. Ware
 * @param <O> the type of object the list will hold
 */
public class DList<O> extends List<O> {

	/**
	 * <p>Construct a dynamic list with a given size and a given first node.</p>
	 * 
	 * @param length the number of elements in the list
	 * @param last the last node in the list
	 */
	protected DList(int length, Node<O> last){
		super(length, last);
	}
	
	/**
	 * <p>Construct an empty dynamic list.</p>
	 */
	public DList(){
		super();
	}
	
	/**
	 * <p>Construct a dynamic list which initially contains a given set of elements.</p>
	 * 
	 * @param elements the elements the list will initially contain
	 */
	public DList(O[] elements){
		super(elements);
	}
	
	/**
	 * <p>Return the number of elements in the dynamic list in O(1) time.</p>
	 * 
	 * @return the number of elements
	 */
	public synchronized int length(){
		return super.length();
	}
	
	/**
	 * <p>Add an element to the dynamic list in O(1) time.</p>
	 * <p>The first element (index 0) is added first, which means that calling {@link #getLast} on the returned list will return <code>elements[elements.length - 1]</code>.<p>
	 * 
	 * @param element the element to be added
	 * @return this list
	 */
	public synchronized DList<O> add(O element){
		last = new Node<O>(element, last);
		length++;
		return this;
	}
	
	/**
	 * <p>Add x elements to the dynamic list in O(x) time.</p>
	 * 
	 * @param elements an array of elements to be added
	 * @return this list
	 */
	public synchronized DList<O> add(O[] elements){
		for(int i=0; i<elements.length; i++)
			last = new Node<O>(elements[i], last);
		length += elements.length;
		return this;
	}
	
	/**
	 * <p>Add every element from <tt>otherList</tt> to the dynamic list.</p>
	 * <p>The first element added is the first element that was added to <tt>otherList</tt>, which means <code>result.{@link #getLast}==otherList.{@link #getLast}</code>.</p>
	 * 
	 * @param otherList a list of elements to be added to this list
	 * @return this list
	 */
	public synchronized DList<O> add(List<O> otherList){
		last = recursiveConcatenate(otherList.last);
		length += otherList.length();
		return this;
	}
	
	/**
	 * <p>A recursive helper function for {@link #add(List)} that adds the first element of <tt>otherList</tt> first.</p>
	 * 
	 * @param node the current node in the list to be copied
	 * @return a new chain of nodes containing the elements of <tt>otherList</tt> in order and ending in {@link #first}.
	 */
	private final Node<O> recursiveConcatenate(Node<O> node){
		if(node.next == null)
			return new Node<O>(node.content, last);
		else
			return new Node<O>(node.content, recursiveConcatenate(node.next));
	}
	
	/**
	 * <p>Return the last element added to the dynamic list in O(1) time.</p>
	 * 
	 * @return the last element
	 */
	public synchronized O getLast(){
		return super.getLast();
	}
	
	/**
	 * <p>Get an arbitrary element from the dynamic list in O(n) time.<p>
	 * <p><code>get(0)</code> is equivalent to {@link #getLast}.</p>
	 * 
	 * @param index the number of the element to be returned
	 * @return the <tt>index</tt>'th element or null if <tt>index</tt> is greater than or equal to {@link #length}.
	 */
	public synchronized O get(int index){
		return super.get(index);
	}
	
	/**
	 * <p>Search the dynamic list to see if it contains a given object in O(n) time.  Objects are compared using <code>==</code>.</p>
	 * 
	 * @param element the object to search for
	 * @return <tt>true</tt> is the object is in this list, <tt>false</tt> otherwise
	 */
	public synchronized boolean contains(O element){
		return super.contains(element);
	}
	
	/**
	 * <p>Search the dynamic list to see if it contains a given object in O(n) time.  Objects are compared using {@link Object#equals}.</p>
	 * 
	 * @param element the object to search for
	 * @return <tt>true</tt> is the object is in this list, <tt>false</tt> otherwise
	 */
	public synchronized boolean containsEquals(O element){
		return super.containsEquals(element);
	}
	
	/**
	 * <p>Remove the first instance of a given object from the dynamic list in O(n) time.  Objects are compared using <code>==</code>.</p>
	 * 
	 * @param element the object to remove
	 * @return this list
	 */
	public synchronized DList<O> remove(O element){
		if(length == 0)
			return this;
		else if(last.content == element){
			last = last.next;
			length--;
			return this;
		}
		else{
			Node<O> current = last;
			while(current.next != null){
				if(current.next.content == element){
					current.next = current.next.next;
					length--;
					return this;
				}
				current = current.next;
			}
			return this;
		}
	}
	
	/**
	 * <p>Remove the first instance of a given object from the dynamic list in O(n) time.  Objects are compared using {@link Object#equals}.</p>
	 * 
	 * @param element the object to remove
	 * @return this list
	 */
	public synchronized DList<O> removeEquals(O element){
		if(length == 0)
			return this;
		else if(last.content.equals(element)){
			last = last.next;
			length--;
			return this;
		}
		else{
			Node<O> current = last;
			while(current.next != null){
				if(current.next.content.equals(element)){
					current.next = current.next.next;
					length--;
					return this;
				}
				current = current.next;
			}
			return this;
		}
	}
	
	/**
	 * <p>Remove every instance of a given object from the dynamic list in O(n) time.  Objects are compared using <code>==</code>.</p>
	 * 
	 * @param element the object to remove every instance of
	 * @return this list
	 */
	public synchronized DList<O> removeAll(O element){
		while(last != null && last.content == element){
			last = last.next;
			length--;
		}
		if(length == 0)
			return this;
		Node<O> current = last;
		while(current.next != null){
			if(current.next.content == element){
				current.next = current.next.next;
				length--;
			}
			else current = current.next;
		}
		return this;
	}
	
	/**
	 * <p>Remove every instance of a given object from the dynamic list in O(n) time.  Objects are compared using {@link Object#equals}.</p>
	 * 
	 * @param element the object to remove every instance of
	 * @return this list
	 */
	public synchronized DList<O> removeAllEquals(O element){
		while(last != null && last.content.equals(element)){
			last = last.next;
			length--;
		}
		if(length == 0)
			return this;
		Node<O> current = last;
		while(current.next != null){
			if(current.next.content.equals(element)){
				current.next = current.next.next;
				length--;
			}
			else current = current.next;
		}
		return this;
	}
	
	/**
	 * <p>Return an {@link java.util.Iterator} that will yield each element in the dynamic list.<p>
	 * <p>The first element yielded will be the last element added to the list; the last element yielded will be the first element added to the list.</p>
	 * 
	 * @return an iterator
	 */
	public synchronized Iterator<O> iterator(){
		return super.iterator();
	}
	
	/**
	 * <p>Return an array containing all the elements in the dynamic list.</p>
	 * <p>The first element in the array (index 0) will be the first element added to the list.</p>
	 * 
	 * @param array an array of type O of any length
	 * @return an array of all the elements in this list
	 */
	public synchronized O[] toArray(O[] array){
		return super.toArray(array);
	}
	
	/**
	 * <p>Check if this list is the same as another.</p>
	 * <p>Two lists are considered the same if they contain the same elements in the same order.  Elements are compared using {@link Object#equals}.</p>
	 * <p>The class of the list is not taken into account.  In other words, a {@link DList} can be equal to an {@link SList} if they contains the same elements in the same order.</p>
	 * 
	 * @param other another list
	 * @return <tt>true</tt> is the lists are the same, <tt>false</tt> if the lists are different or if <tt>other</tt> is not a list
	 */
	public synchronized boolean equals(Object other){
		return super.equals(other);
	}
	
	/**
	 * <p>Return a deep copy of the dynamic list.</p>
	 * <p>The clone list can be modified without affecting the original.</p>
	 */
	public synchronized DList<O> clone(){
		return new DList<O>(length, last.clone());
	}
	
	/**
	 * <p>Return a string representation of this dynamic list and all its elements.</p>
	 */
	public synchronized String toString(){
		return super.toString();
	}
}
