package sdds;

import java.util.Iterator;

/**
 * <p>A simple hash-table (the abstract superclass of {@link DHashTable} and {@link SHashTable}).</p>
 * <p>The performance of each operation varies for {@link DHashTable} and {@link SHashTable},
 * but in both cases searching for a value based on its key takes O(1) amortized time (assuming the hash codes of keys are uniformly distributed).</p>
 * <p>Note that key/value mappings can be added to the hash table using <code>==</code> or {@link Object#equals}.
 * Using {@link #add} and {@link #addEquals} on the same table may put it in an inconsistent state.
 * For example, say that K1 is mapped to V1 using {@link #add} and K2 is mapped to V2 using {@link #add}.
 * If K1.equals(K2), remapping K1 using {@link #addEquals} could overwrite either one of the above mappings.</p>
 * 
 * @author Stephen G. Ware
 * @param <K> the type of object that will act as keys for the hash table
 * @param <V> the type of object that is being stored in the hash table
 */
public abstract class HashTable<K, V> implements Cloneable {

	/** The default number of mapping that can exist in a new hash table */
	protected static final int DEFAULT_INITIAL_SIZE = 10;
	
	/** The default number of times larger that the hash table will get when it needs to expand */
	protected static final int DEFAULT_EXPANSION_FACTOR = 2;
	
	/** The load factor at or above which the hash table will expand */
	protected static final float DEFAULT_LOAD_FACTOR_THRESHOLD = .75f;
	
	/**
	 * <p>Each bucket in the hash table stores one of these nodes, which contains a key/value mapping and a pointer to the next node in the chain.</p>
	 * 
	 * @author Stephen G. Ware
	 * @param <NK> the type of object that will act as keys for the hash table
	 * @param <NV> the type of object that is being stored in the hash table
	 */
	static class Node<NK, NV> {
		
		/** The key object for this mapping */
		public NK key;
		
		/** The item being mapped to */
		public NV value;
		
		/** The next node in the chain */
		public Node<NK, NV> next;
		
		/**
		 * <p>A node is constructed with a key/value mapping and a pointer to the next node in the chain.</p>
		 * 
		 * @param key the key for this mapping
		 * @param value the object being mapped to
		 * @param next the next node in the chain
		 */
		public Node(NK key, NV value, Node<NK, NV> next){
			this.key = key;
			this.value = value;
			this.next = next;
		}
		
		/**
		 * <p>Return a copy of the mapping and all nodes later in the chain.</p>
		 */
		public Node<NK, NV> clone(){
			if(next == null)
				return new Node<NK, NV>(key, value, null);
			else
				return new Node<NK, NV>(key, value, next.clone());
		}
	}

	/** The number of mapping in this hash table */
	protected int size;
	
	/** The buckets */
	protected Node<K, V>[] table;
	
	/**
	 * <p>Construct a hash table with a given number of mappings and given set of buckets.</p>
	 * 
	 * @param size the number of mappings in the hash table
	 * @param table the buckets
	 */
	protected HashTable(int size, Node<K, V>[] table){
		this.size = size;
		this.table = table;
	}

	/**
	 * <p>Construct an empty hash table.</p>
	 */
	@SuppressWarnings("unchecked")
	public HashTable(){
		this(0, (Node<K, V>[]) new Node[DEFAULT_INITIAL_SIZE]);
	}
	
	/**
	 * <p>Construct a hash table which initially contains mappings for a given set of keys and values.</p>
	 * <p>Mappings are added as if using {@link #add}.</p>
	 * 
	 * @param keys the objects which can be used to retrieve the values
	 * @param values the objects to be retrieved using the keys
	 * @throws ArrayIndexOutOfBoundsException if the array of keys is not the same length as the array of values
	 */
	@SuppressWarnings("unchecked")
	public HashTable(K[] keys, V[] values){
		if(keys.length != values.length)
			throw new ArrayIndexOutOfBoundsException("The array of keys must be the same length as the array of values.");
		table = (Node<K, V>[]) new Node[(int) (((float) keys.length) / DEFAULT_LOAD_FACTOR_THRESHOLD)];
		size = 0;
		int bucket;
		Node<K, V> current;
		boolean newMapping;
		for(int i=0; i<keys.length; i++){
			bucket = getBucket(keys[i], table);
			newMapping = true;
			current = table[bucket];
			while(current != null){
				if(current.key == keys[i]){
					current.key = keys[i];
					current.value = values[i];
					newMapping = false;
					break;
				}
				current = current.next;
			}
			if(newMapping){
				table[bucket] = new Node<K, V>(keys[i], values[i], table[bucket]);
				size++;
			}
		}
	}
	
	/**
	 * <p>Given an object that needs to be placed in a bucket, this function return the number of the bucket in which to place the object.</p>
	 * 
	 * @param obj the object to place
	 * @param table the buckets
	 * @return the index of the bucket to use
	 */
	protected static final int getBucket(Object obj, Node<?, ?>[] table){
		int b = obj.hashCode() % table.length;
		if(b < 0) return b * -1;
		else return b;
	}
	
	/**
	 * <p>Return the load factor of the hash table.</p>
	 * <p>The load factor is defined as the number of mapping divided by the number of buckets.</p>
	 * 
	 * @return the load factor
	 */
	protected float getLoadFactor(){
		return ((float) size) / table.length;
	}
	
	/**
	 * <p>Expand the buckets into <code>buckets</code> new buckets.  Re-map all existing mappings to their new buckets.</p>
	 * 
	 * @param buckets the new number of buckets
	 * @return the new buckets
	 */
	@SuppressWarnings("unchecked")
	protected final Node<K, V>[] rehash(int buckets){
		Node<K, V>[] newTable = (Node<K, V>[]) new Node[buckets];
		Node<K, V> current;
		int bucket;
		for(int i=0; i<table.length; i++){
			current = table[i];
			while(current != null){
				bucket = getBucket(current.key, newTable);
				newTable[bucket] = new Node<K, V>(current.key, current.value, newTable[bucket]);
				current = current.next;
			}
		}
		return newTable;
	}
	
	/**
	 * <p>Expand the buckets into the number of current buckets times {@link #DEFAULT_EXPANSION_FACTOR}.  Re-map all existing mappings to their new buckets.</p>
	 * 
	 * @return the new buckets
	 */
	protected final Node<K, V>[] rehash(){
		return rehash(table.length * DEFAULT_EXPANSION_FACTOR);
	}
	
	/**
	 * <p>Return the number of mappings in the hash table.</p>
	 * 
	 * @return the number of mappings
	 */
	public int size(){
		return size;
	}
	
	/**
	 * <p>Add a key/value mapping to the hash table.  If a mapping exists whose key is <code>==</code> to this key, it is replaced.</p>
	 * 
	 * @param key the object which can be used to retrieve the value
	 * @param value the object to be retrieved using the key
	 * @return a hash table exactly like this one, but with the mapping added
	 */
	public abstract HashTable<K, V> add(K key, V value);
	
	/**
	 * <p>Add multiple key/value mappings to the hash table.  If mappings exist whose keys are <code>==</code> to the given keys, they are replaced.</p>
	 * 
	 * @param keys the objects which can be used to retrieve the values
	 * @param values the objects to be retrieved using the keys
	 * @return a hash table exactly like this one, but with the mappings added
	 * @throws ArrayIndexOutOfBoundsException if the array of keys is not the same length as the array of values
	 */
	public abstract HashTable<K, V> add(K[] keys, V[] values);
	
	/**
	 * <p>Add a key/value mapping to the hash table.  If a mapping exists whose key {@link Object#equals} this key, it is replaced.</p>
	 * 
	 * @param key the object which can be used to retrieve the value
	 * @param value the object to be retrieved using the key
	 * @return a hash table exactly like this one, but with the mapping added
	 */
	public abstract HashTable<K, V> addEquals(K key, V value);
	
	/**
	 * <p>Add multiple key/value mappings to the hash table.  If mappings exist whose keys are {@link Object#equals} to the given keys, they are replaced.</p>
	 * 
	 * @param keys the objects which can be used to retrieve the values
	 * @param values the objects to be retrieved using the keys
	 * @return a hash table exactly like this one, but with the mappings added
	 * @throws ArrayIndexOutOfBoundsException if the array of keys is not the same length as the array of values
	 */
	public abstract HashTable<K, V> addEquals(K[] keys, V[] values);
	
	/**
	 * <p>Return the value mapped to a given key.  Keys are searched using <code>==</code>.</p>
	 * 
	 * @param key the object that was mapped to the desired value
	 * @return the value
	 */
	public V get(K key){
		int bucket = getBucket(key, table);
		Node<K, V> current = table[bucket];
		while(current != null){
			if(current.key == key)
				return current.value;
			current = current.next;
		}
		return null;
	}
	
	/**
	 * <p>Return the value mapped to a given key.  Keys are searched using {@link Object#equals}.</p>
	 * 
	 * @param key the object that was mapped to the desired value
	 * @return the value
	 */
	public V getEquals(K key){
		int bucket = getBucket(key, table);
		Node<K, V> current = table[bucket];
		while(current != null){
			if(current.key.equals(key))
				return current.value;
			current = current.next;
		}
		return null;
	}
	
	/**
	 * <p>Remove a key/value mapping from the hash table.  Keys are searched using <code>==</code>.</p>
	 * 
	 * @param key the key of the key/value mapping to be removed
	 * @return a hash table exactly like this one, but with the mapping removed
	 */
	public abstract HashTable<K, V> remove(K key);
	
	/**
	 * <p>Remove multiple key/value mappings from the hash table.  Keys are searched using <code>==</code>.</p>
	 * 
	 * @param keys the keys of the key/value mappings to be removed
	 * @return a hash table exactly like this one, but with the mappings removed
	 */
	public abstract HashTable<K, V> remove(K[] keys);
	
	/**
	 * <p>Remove a key/value mapping from the hash table.  Keys are searched using {@link Object#equals}.</p>
	 * 
	 * @param key the key of the key/value mapping to be removed
	 * @return a hash table exactly like this one, but with the mapping removed
	 */
	public abstract HashTable<K, V> removeEquals(K key);
	
	/**
	 * <p>Remove multiple key/value mappings from the hash table.  Keys are searched using {@link Object#equals}.</p>
	 * 
	 * @param keys the keys of the key/value mappings to be removed
	 * @return a hash table exactly like this one, but with the mappings removed
	 */
	public abstract HashTable<K ,V> removeEquals(K[] keys);
	
	/**
	 * <p>Return an {@link java.util.Iterator} that will yield all the keys in the hash table.</p>
	 * 
	 * @return an iterator of keys
	 */
	public Iterator<K> keys(){
		if(size == 0) return new EmptyIterator<K>();
		else return new HashTableKeyIterator<K, V>(this);
	}
	
	/**
	 * <p>Return an {@link java.util.Iterator} that will yield all the values in the hash table.</p>
	 * 
	 * @return an iterator of values
	 */
	public Iterator<V> values(){
		if(size == 0) return new EmptyIterator<V>();
		else return new HashTableValueIterator<K, V>(this);
	}
	
	/**
	 * <p>Check if this hash table is the same as another.</p>
	 * <p>Two hash tables are considered the same if they contain the same key/value mappings.  Keys and values are compared using {@link Object#equals}.</p>
	 * <p>The class of the hash table is not taken into account.  In other words, a {@link DHashTable} can be equal to an {@link SHashTable} if they contain the same key/value mappings.</p>
	 * 
	 * @param other another hash table
	 * @return <tt>true</tt> is the hash tables are the same, <tt>false</tt> if the hash tables are different or if <tt>other</tt> is not a hash table
	 */
	@SuppressWarnings("unchecked")
	public boolean equals(Object other){
		if(!(other instanceof HashTable<?, ?>))
			return false;
		HashTable<K, V> otherht = (HashTable<K, V>) other;
		if(size != otherht.size || table.length != otherht.table.length)
			return false;
		K key;
		for(Iterator<K> keys = keys(); keys.hasNext();){
			key = keys.next();
			if(!getEquals(key).equals(otherht.getEquals(key)))
				return false;
		}
		return true;
	}

	/**
	 * <p>Return a new hash table which {@link Object#equals} this hash table but is not <code>==</code> to this hash table.</p>
	 */
	public abstract HashTable<K, V> clone();
	
	/**
	 * <p>Clone the buckets, including every link in every chain.</p>
	 * 
	 * @return the cloned buckets
	 */
	@SuppressWarnings("unchecked")
	protected Node<K, V>[] cloneTable(){
		Node<K, V>[] newTable = (Node<K, V>[]) new Node[table.length];
		for(int i=0; i<table.length; i++)
			if(table[i] != null)
				newTable[i] = table[i].clone();
		return newTable;
	}
	
	/**
	 * <p>Return a string representation of this hash table and all its elements.</p>
	 */
	public String toString(){
		String str = this.getClass().getSimpleName() + ":";
		Node<K, V> current;
		for(int i=0; i<table.length; i++){
			str += "\n[bucket " + i + "] ";
			current = table[i];
			while(current != null){
				str += current.key + " => " + current.value + " ";
				current = current.next;
			}
		}
		return str;
	}
}
