package sdds;

import sdds.HashTable.Node;

/**
 * <p>Iterates through each key/value mapping in a {@link HashTable}.
 * This is the abstract superclass of {@link HashTableKeyIterator} and {@link HashTableValueIterator}.</p>
 * 
 * @author Stephen G. Ware
 * @param <K> the key object for the hash table
 * @param <V> the value object for the hash table
 */
abstract class HashTableIterator<K, V> {
	
	/** The buckets of the hash table */
	protected final Node<K, V>[] table;

	/** The next bucket to search */
	protected int next = 0;
	
	/** The current key/value mapping */
	protected Node<K, V> current;
	
	/**
	 * <p>Construct an iterator for a given {@link HashTable}.</p>
	 * 
	 * @param hashTable
	 */
	protected HashTableIterator(HashTable<K, V> hashTable){
		this.table = hashTable.table;
		next = -1;
		do{
			next++;
			current = table[next];
		} while(current == null);
	}
	
	/**
	 * <p>Test if the hash table has any more element to yield.</p>
	 * 
	 * @return <tt>true</tt> if there are more elements, <tt>false</tt> otherwise
	 */
	public boolean hasNext(){
		return current != null;
	}
	
	/**
	 * <p>Finds the next key/value mapping in the hash table.</p>
	 * 
	 * @return the next mapping
	 */
	protected Node<K, V> getNextNode(){
		if(current == null)
			throw new java.util.NoSuchElementException("No more elements exists for this iterator.");
		Node<K, V> nextNode = current;
		current = current.next;
		while(current == null && next < table.length - 1){
			next++;
			current = table[next];
		}
		return nextNode;
	}
	
	/**
	 * <p>Always throws a {@link UnsupportedOperationException}.</p>
	 * 
	 * @throws UnsupportedOperationException every time the method is called
	 */
	public void remove(){
		throw new UnsupportedOperationException("This iterator does not support the remove method.");
	}
}
