package sdds;

import java.util.Iterator;

/**
 * <p>A singly-linked list (the abstract superclass of {@link DList} and {@link SList}).<p>
 * <p>In general, the time for each operation is as follows:</p>
 * <ul>
 * <li>add an element: O(1) time</li>
 * <li>search for an element: O(n) time</li>
 * <li>remove the last element: O(1) time</li>
 * <li>remove any element: O(n) time</li>
 * </ul>
 * 
 * @author Stephen G. Ware
 * @param <O> the type of object the list will hold
 */
public abstract class List<O> {
	
	/**
	 * <p>This object represents an individual node in the linked list.</p>
	 * 
	 * @author Stephen G. Ware
	 * @param <NO> the type of object the node will hold
	 */
	static class Node<NO> {
		
		/** The element of the list */
		public final NO content;
		
		/** The next element in the list */
		public Node<NO> next;
		
		/**
		 * <p>A node holds an element of the list and a pointer to the next node in the list.<p>
		 * 
		 * @param content the element of the list
		 * @param next a pointer to the next node
		 */
		public Node(NO content, Node<NO> next){
			this.content = content;
			this.next = next;
		}
		
		/**
		 * <p>Recursively copy this node and every subsequent node.</p>
		 * 
		 * @return a clone of the node and all subsequent nodes
		 */
		public Node<NO> clone(){
			if(next == null)
				return new Node<NO>(content, null);
			else
				return new Node<NO>(content, next.clone());
		}
	}
	
	/** The number of elements in the list */
	protected int length;
	
	/** The last node in the list */
	protected Node<O> last;

	/**
	 * <p>Construct a list with a given size and a given first node.</p>
	 * 
	 * @param length the number of elements in the list
	 * @param last the first node in the list
	 */
	protected List(int length, Node<O> last){
		this.length = length;
		this.last = last;
	}
	
	/**
	 * <p>Construct an empty list.</p>
	 */
	protected List(){
		this(0, null);
	}
	
	/**
	 * <p>Construct a list which initially contains a given set of elements.</p>
	 * 
	 * @param elements the elements the list will initially contain
	 */
	protected List(O[] elements){
		for(int i=0; i<elements.length; i++)
			last = new Node<O>(elements[i], last);
		length = elements.length;
	}
	
	/**
	 * <p>Return the number of elements in the list in O(1) time.</p>
	 * 
	 * @return the number of elements
	 */
	public int length(){
		return length;
	}

	/**
	 * <p>Add an element to the list in O(1) time.</p>
	 * <p>The first element (index 0) is added first, which means that calling {@link #getLast} on the returned list will return <code>elements[elements.length - 1]</code>.<p>
	 * 
	 * @param element the element to be added
	 * @return a list exactly like this one, but with <tt>element</tt> added to it
	 */
	public abstract List<O> add(O element);
	
	/**
	 * <p>Add x elements to the list in O(x) time.</p>
	 * 
	 * @param elements an array of elements to be added
	 * @return a list exactly like this one, but with all the <tt>elements</tt> added to it
	 */
	public abstract List<O> add(O[] elements);
	
	/**
	 * <p>Add every element from <tt>otherList</tt> to the list.</p>
	 * <p>The first element added is the first element that was added to <tt>otherList</tt>, which means <code>result.{@link #getLast}==otherList.{@link #getLast}</code>.</p>
	 * 
	 * @param otherList a list of elements to be added to this list
	 * @return a list exactly like this one, but with each element from <tt>otherList</tt> added to it
	 */
	public abstract List<O> add(List<O> otherList);
	
	/**
	 * <p>Return the last element added to the list in O(1) time.</p>
	 * 
	 * @return the last element
	 */
	public O getLast(){
		if(last == null) return null;
		else return last.content;
	}
	
	/**
	 * <p>Get an arbitrary element from the list in O(n) time.<p>
	 * <p><code>get(list.length() - 1)</code> is equivalent to {@link #getLast}.</p>
	 * 
	 * @param index the number of the element to be returned
	 * @return the <tt>index</tt>'th element or null if <tt>index</tt> is greater than or equal to {@link #length}.
	 */
	public O get(int index){
		if(index >= length) return null;
		Node<O> current = last;
		index = length - index - 1;
		for(int i=0; i<index; i++)
			current = current.next;
		return current.content;
	}
	
	/**
	 * <p>Search the list to see if it contains a given object in O(n) time.  Objects are compared using <code>==</code>.</p>
	 * 
	 * @param element the object to search for
	 * @return <tt>true</tt> is the object is in this list, <tt>false</tt> otherwise
	 */
	public boolean contains(O element){
		Node<O> current = last;
		while(current != null){
			if(current.content == element)
				return true;
			current = current.next;
		}
		return false;
	}
	
	/**
	 * <p>Search the list to see if it contains a given object in O(n) time.  Objects are compared using {@link Object#equals}.</p>
	 * 
	 * @param element the object to search for
	 * @return <tt>true</tt> is the object is in this list, <tt>false</tt> otherwise
	 */
	public boolean containsEquals(O element){
		Node<O> current = last;
		while(current != null){
			if(current.content.equals(element))
				return true;
			current = current.next;
		}
		return false;
	}
	
	/**
	 * <p>Remove the first instance of a given object from the list in O(n) time.  Objects are compared using <code>==</code>.</p>
	 * 
	 * @param element the object to remove
	 * @return a list exactly like this one, but which contains one less instance of <tt>element</tt>, or the same list if no instance of <tt>element</tt> was found
	 */
	public abstract List<O> remove(O element);
	
	/**
	 * <p>Remove the first instance of a given object from the list in O(n) time.  Objects are compared using {@link Object#equals}.</p>
	 * 
	 * @param element the object to remove
	 * @return a list exactly like this one, but which contains one less instance of <tt>element</tt>, or the same list if no instance of <tt>element</tt> was found
	 */
	public abstract List<O> removeEquals(O element);
	
	/**
	 * <p>Remove every instance of a given object from the list in O(n) time.  Objects are compared using <code>==</code>.</p>
	 * 
	 * @param element the object to remove every instance of
	 * @return a list exactly like this one, but which contains no instances of <tt>element</tt>, or the same list if no instances of <tt>element</tt> were found
	 */
	public abstract List<O> removeAll(O element);
	
	/**
	 * <p>Remove every instance of a given object from the list in O(n) time.  Objects are compared using {@link Object#equals}.</p>
	 * 
	 * @param element the object to remove every instance of
	 * @return a list exactly like this one, but which contains no instances of <tt>element</tt>, or the same list if no instances of <tt>element</tt> were found
	 */
	public abstract List<O> removeAllEquals(O element);
	
	/**
	 * <p>Return an {@link java.util.Iterator} that will yield each element in the list.<p>
	 * <p>The first element yielded will be the last element added to the list; the last element yielded will be the first element added to the list.</p>
	 * 
	 * @return an iterator
	 */
	public Iterator<O> iterator(){
		return new ListIterator<O>(this);
	}
	
	/**
	 * <p>Return an array containing all the elements in this list.</p>
	 * <p>The first element in the array (index 0) will be the first element added to the list.</p>
	 * 
	 * @param array an array of type O of any length
	 * @return an array of all the elements in this list
	 */
	public O[] toArray(O[] array){
		array = java.util.Arrays.copyOf(array, length);
		Node<O> current = last;
		for(int i = length - 1; i>=0; i--){
			array[i] = current.content;
			current = current.next;
		}
		return array;
	}
	
	/**
	 * <p>Check if this list is the same as another.</p>
	 * <p>Two lists are considered the same if they contain the same elements in the same order.  Elements are compared using {@link Object#equals}.</p>
	 * <p>The class of the list is not taken into account.  In other words, a {@link DList} can be equal to an {@link SList} if they contain the same elements in the same order.</p>
	 * 
	 * @param other another list
	 * @return <tt>true</tt> is the lists are the same, <tt>false</tt> if the lists are different or if <tt>other</tt> is not a list
	 */
	@SuppressWarnings("unchecked")
	public boolean equals(Object other){
		if(!(other instanceof List<?>))
			return false;
		List<O> otherList = (List<O>) other;
		if(length != otherList.length)
			return false;
		Node<O> myCurrent = last;
		Node<O> otherCurrent = otherList.last;
		while(myCurrent != null && otherCurrent != null){
			if(!myCurrent.content.equals(otherCurrent.content))
				return false;
			myCurrent = myCurrent.next;
			otherCurrent = otherCurrent.next;
		}
		return true;
	}

	/**
	 * <p>Return a new list which {@link Object#equals} this list but is not <code>==</code> to this list.</p>
	 */
	public abstract List<O> clone();
	
	/**
	 * <p>Return a string representation of this list and all its elements.</p>
	 */
	@SuppressWarnings("unchecked")
	public String toString(){
		String str = getClass().getSimpleName() + ": ";
		O[] elements = toArray((O[]) new Object[0]);
		boolean first = true;
		for(O element : elements){
			if(first) first = false;
			else str += ", ";
			str += element;
		}
		return str;
	}
}
