package sdds;

/**
 * <p>A simple static {@link HashTable}.</p>
 * <p>In general, the time for each operation is as follows:</p>
 * <ul>
 * <li>add an element: O(n) time (because a new set of buckets must be allocated)</li>
 * <li>search for an element: O(1) amortized time</li>
 * <li>remove an element: O(n) time (because a new set of buckets must be allocated)</li>
 * </ul>
 * 
 * @author Stephen G. Ware
 * @param <K> the type of object that will act as keys for the hash table
 * @param <V> the type of object that is being stored in the hash table
 */
public class SHashTable<K, V> extends HashTable<K, V> {
	
	/**
	 * <p>Construct a static hash table with a given number of mappings and given set of buckets.</p>
	 * 
	 * @param size the number of mappings in the hash table
	 * @param table the buckets
	 */
	protected SHashTable(int size, Node<K, V>[] table){
		super(size, table);
	}

	/**
	 * <p>This protected method is used instead of a constructor to create new hash tables when this hash table is modified.</p>
	 * <p>All subclasses of SHashTable should use this method instead of a constructor.
	 * This method can be overridden in a subclass and made to return objects of the subclass's type.
	 * This avoids the new to re-write the modifying methods such as {@link #add} and {@link #remove} in every subclass.</p>
	 * 
	 * @param size the number of mappings in the hash table
	 * @param table the buckets
	 */
	protected SHashTable<K, V> makeHashTable(int size, Node<K, V>[] table){
		return new SHashTable<K, V>(size, table);
	}
	
	/**
	 * <p>Construct an empty static hash table.</p>
	 */
	public SHashTable(){
		super();
	}

	/**
	 * <p>Construct a static hash table which initially contains mappings for a given set of keys and values.</p>
	 * <p>Mappings are added as if using {@link #add}.</p>
	 * <p>This constructor is more efficient in both time and space than creating an empty static hash table and then adding the key/value mappings.</p>
	 * 
	 * @param keys the objects which can be used to retrieve the values
	 * @param values the objects to be retrieved using the keys
	 * @throws ArrayIndexOutOfBoundsException if the array of keys is not the same length as the array of values
	 */
	public SHashTable(K[] keys, V[] values){
		super(keys, values);
	}
	
	/**
	 * <p>Add a key/value mapping to the static hash table.  If a mapping exists whose key is <code>==</code> to this key, it is replaced.</p>
	 * 
	 * @param key the object which can be used to retrieve the value
	 * @param value the object to be retrieved using the key
	 * @return a clone of the hash table with the mapping added
	 */
	public SHashTable<K, V> add(K key, V value){
		Node<K, V>[] newTable;
		if(getLoadFactor() > DEFAULT_LOAD_FACTOR_THRESHOLD) newTable = rehash();
		else newTable = table.clone();
		int bucket = getBucket(key, newTable);
		Node<K, V> newChain = copyChainReplace(newTable[bucket], key, value);
		if(newChain == newTable[bucket]){
			newTable[bucket] = new Node<K, V>(key, value, newTable[bucket]);
			return makeHashTable(size + 1, newTable);
		}
		else{
			newTable[bucket] = newChain;
			return makeHashTable(size, newTable);
		}
	}

	/**
	 * <p>A recursive helper function for {@link #add}.  If the given mapping exists, it is replaced and the chain copied.
	 * Otherwise, the original chain is returned.</p>
	 * 
	 * @param node the current node in the chain
	 * @param key the key to replace
	 * @param value the value to replace
	 * @return a new chain with the mapping replaced or the original chain
	 */
	private final Node<K, V> copyChainReplace(Node<K, V> node, K key, V value){
		if(node == null) return null;
		else if(node.key == key) return new Node<K, V>(key, value, node.next);
		else{
			Node<K, V> newChain = copyChainReplace(node.next, key, value);
			if(newChain == node.next) return node;
			else return new Node<K, V>(node.key, node.value, newChain);
		}
	}
	
	/**
	 * <p>Add multiple key/value mappings to the static hash table.  If mappings exist whose keys are <code>==</code> to the given keys, they are replaced.</p>
	 * <p>This method is more efficient in both time and space than making multiple calls to {@link #add}.</p>
	 * 
	 * @param keys the objects which can be used to retrieve the values
	 * @param values the objects to be retrieved using the keys
	 * @return a clone of the hash table with the mappings added
	 * @throws ArrayIndexOutOfBoundsException if the array of keys is not the same length as the array of values
	 */
	public SHashTable<K, V> add(K[] keys, V[] values){
		if(keys.length != values.length)
			throw new ArrayIndexOutOfBoundsException("The array of keys must be the same length as the array of values.");
		if(keys.length == 0) return this;
		// Add every key value pair.
		int newSize = size;
		Node<K, V>[] newTable = table.clone();
		int bucket;
		Node<K, V> newChain;
		for(int i=0; i<keys.length; i++){
			bucket = getBucket(keys[i], newTable);
			newChain = copyChainReplace(newTable[bucket], keys[i], values[i]);
			if(newChain == newTable[bucket]){
				newTable[bucket] = new Node<K, V>(keys[i], values[i], newTable[bucket]);
				newSize++;
			}
			else
				newTable[bucket] = newChain;
		}
		// Rehash the table if needed.
		SHashTable<K, V> newHashTable = makeHashTable(newSize, newTable);
		if(newHashTable.getLoadFactor() > DEFAULT_LOAD_FACTOR_THRESHOLD)
			newHashTable.table = newHashTable.rehash((int) (((float) newTable.length) / DEFAULT_LOAD_FACTOR_THRESHOLD));
		return newHashTable;
	}
	
	/**
	 * <p>Add a key/value mapping to the static hash table.  If a mapping exists whose key {@link Object#equals} this key, it is replaced.</p>
	 * 
	 * @param key the object which can be used to retrieve the value
	 * @param value the object to be retrieved using the key
	 * @return a clone of the hash table with the mapping added
	 */
	public SHashTable<K, V> addEquals(K key, V value){
		Node<K, V>[] newTable;
		if(getLoadFactor() > DEFAULT_LOAD_FACTOR_THRESHOLD) newTable = rehash();
		else newTable = table.clone();
		int bucket = getBucket(key, newTable);
		Node<K, V> newChain = copyChainReplaceEquals(newTable[bucket], key , value);
		if(newChain == newTable[bucket]){
			newTable[bucket] = new Node<K, V>(key, value, newTable[bucket]);
			return makeHashTable(size + 1, newTable);
		}
		else{
			newTable[bucket] = newChain;
			return makeHashTable(size, newTable);
		}
	}
	
	/**
	 * <p>A recursive helper function for {@link #addEquals}.  If the given mapping exists, it is replaced and the chain copied.
	 * Otherwise, the original chain is returned.</p>
	 * 
	 * @param node the current node in the chain
	 * @param key the key to replace
	 * @param value the value to replace
	 * @return a new chain with the mapping replaced or the original chain
	 */
	private final Node<K, V> copyChainReplaceEquals(Node<K, V> node, K key, V value){
		if(node == null) return null;
		else if(node.key.equals(key)) return new Node<K, V>(key, value, node.next);
		else{
			Node<K, V> newChain = copyChainReplaceEquals(node.next, key, value);
			if(newChain == node.next) return node;
			else return new Node<K, V>(node.key, node.value, newChain);
		}
	}
	
	/**
	 * <p>Add multiple key/value mappings to the static hash table.  If mappings exist whose keys are {@link Object#equals} to the given keys, they are replaced.</p>
	 * <p>This method is more efficient in both time and space than multiple calls to {@link #addEquals}.</p>
	 * 
	 * @param keys the objects which can be used to retrieve the values
	 * @param values the objects to be retrieved using the keys
	 * @return a clone of the hash table with the mappings added
	 * @throws ArrayIndexOutOfBoundsException if the array of keys is not the same length as the array of values
	 */
	public SHashTable<K, V> addEquals(K[] keys, V[] values){
		if(keys.length != values.length)
			throw new ArrayIndexOutOfBoundsException("The array of keys must be the same length as the array of values.");
		if(keys.length == 0) return this;
		// Add every key value pair.
		int newSize = size;
		Node<K, V>[] newTable = table.clone();
		int bucket;
		Node<K, V> newChain;
		for(int i=0; i<keys.length; i++){
			bucket = getBucket(keys[i], newTable);
			newChain = copyChainReplaceEquals(newTable[bucket], keys[i], values[i]);
			if(newChain == newTable[bucket]){
				newTable[bucket] = new Node<K, V>(keys[i], values[i], newTable[bucket]);
				newSize++;
			}
			else
				newTable[bucket] = newChain;
		}
		// Rehash the table if needed.
		SHashTable<K, V> newHashTable = makeHashTable(newSize, newTable);
		if(newHashTable.getLoadFactor() > DEFAULT_LOAD_FACTOR_THRESHOLD)
			newHashTable.table = newHashTable.rehash((int) (((float) newTable.length) / DEFAULT_LOAD_FACTOR_THRESHOLD));
		return newHashTable;
	}

	/**
	 * <p>Remove a key/value mapping from the static hash table.  Keys are searched using <code>==</code>.</p>
	 * 
	 * @param key the key of the key/value mapping to be removed
	 * @return a clone of the hash table with the mapping removed
	 */
	public SHashTable<K, V> remove(K key){
		int bucket = getBucket(key, table);
		Node<K, V> newChain = copyChainRemove(table[bucket], key);
		if(newChain == table[bucket])
			return this;
		else{
			Node<K, V>[] newTable = table.clone();
			newTable[bucket] = newChain;
			return makeHashTable(size - 1, newTable);
		}
	}
	
	/**
	 * <p>A recursive helper function for {@link #remove}.  If the given mapping exists, a copy of the chain without that mapping is returned.
	 * Otherwise, the original chain is returned.</p>
	 * 
	 * @param node the current node in the chain
	 * @param key the key to remove
	 * @return a new chain with the mapping removed or the original chain
	 */
	private final Node<K, V> copyChainRemove(Node<K, V> node, K key){
		if(node == null) return null;
		else if(node.key == key) return node.next;
		else{
			Node<K, V> newChain = copyChainRemove(node.next, key);
			if(newChain == node.next) return node;
			else return new Node<K, V>(node.key, node.value, newChain);
		}
	}

	/**
	 * <p>Remove multiple key/value mappings from the static hash table.  Keys are searched using <code>==</code>.</p>
	 * <p>This method is more efficient in both time and space than multiple calls to {@link #remove}.</p>
	 * 
	 * @param keys the keys of the key/value mappings to be removed
	 * @return a clone of the hash table with the mappings removed
	 */
	public SHashTable<K, V> remove(K[] keys){
		int bucket;
		Node<K, V> newChain;
		for(int i=0; i<keys.length; i++){
			bucket = getBucket(keys[i], table);
			newChain = copyChainRemove(table[bucket], keys[i]);
			if(newChain != table[bucket]){
				Node<K, V>[] newTable = table.clone();
				newTable[bucket] = newChain;
				int newSize = size - 1;
				for(int j=i+1; j<keys.length; j++){
					bucket = getBucket(keys[i], newTable);
					newChain = copyChainRemove(newTable[bucket], keys[j]);
					if(newChain != newTable[bucket]){
						newTable[bucket] = newChain;
						newSize--;
					}
				}
				return makeHashTable(newSize, newTable);
			}
		}
		return this;
	}

	/**
	 * <p>Remove a key/value mapping from the static hash table.  Keys are searched using {@link Object#equals}.</p>
	 * 
	 * @param key the key of the key/value mapping to be removed
	 * @return a clone of the hash table with the mapping removed
	 */
	public SHashTable<K, V> removeEquals(K key){
		int bucket = getBucket(key, table);
		Node<K, V> newChain = copyChainRemoveEquals(table[bucket], key);
		if(newChain == table[bucket])
			return this;
		else{
			Node<K, V>[] newTable = table.clone();
			newTable[bucket] = newChain;
			return makeHashTable(size - 1, newTable);
		}
	}
	
	/**
	 * <p>A recursive helper function for {@link #removeEquals}.  If the given mapping exists, a copy of the chain without that mapping is returned.
	 * Otherwise, the original chain is returned.</p>
	 * 
	 * @param node the current node in the chain
	 * @param key the key to remove
	 * @return a new chain with the mapping removed or the original chain
	 */
	private final Node<K, V> copyChainRemoveEquals(Node<K, V> node, K key){
		if(node == null) return null;
		else if(node.key.equals(key)) return node.next;
		else{
			Node<K, V> newChain = copyChainRemoveEquals(node.next, key);
			if(newChain == node.next) return node;
			else return new Node<K, V>(node.key, node.value, newChain);
		}
	}

	/**
	 * <p>Remove multiple key/value mappings from the static hash table.  Keys are searched using {@link Object#equals}.</p>
	 * <p>This method is more efficient in both time and space than multiple calls to {@link #removeEquals}.</p>
	 * 
	 * @param keys the keys of the key/value mappings to be removed
	 * @return a hash table exactly like this one, but with the mappings removed
	 */
	public SHashTable<K, V> removeEquals(K[] keys){
		int bucket;
		Node<K, V> newChain;
		for(int i=0; i<keys.length; i++){
			bucket = getBucket(keys[i], table);
			newChain = copyChainRemoveEquals(table[bucket], keys[i]);
			if(newChain != table[bucket]){
				Node<K, V>[] newTable = table.clone();
				newTable[bucket] = newChain;
				int newSize = size - 1;
				for(int j=i+1; j<keys.length; j++){
					bucket = getBucket(keys[i], newTable);
					newChain = copyChainRemoveEquals(newTable[bucket], keys[j]);
					if(newChain != newTable[bucket]){
						newTable[bucket] = newChain;
						newSize--;
					}
				}
				return makeHashTable(newSize, newTable);
			}
		}
		return this;
	}
	
	/**
	 * <p>Return a deep copy of the static hash table.</p>
	 * <p>The clone hash table can be modified without affecting the original.</p>
	 */
	public SHashTable<K, V> clone(){
		return makeHashTable(size, table);
	}
}
