package sdds;

/**
 * <p>A static singly-linked {@link List}.</p>
 * 
 * @author Stephen G. Ware
 * @param <O> the type of object the list will hold
 */
public class SList<O> extends List<O> {
	
	/**
	 * <p>Construct a static list with a given size and a given first node.</p>
	 * 
	 * @param length the number of elements in the list
	 * @param first the first node in the list
	 */
	protected SList(int length, Node<O> first){
		super(length, first);
	}

	/**
	 * <p>This protected method is used instead of a constructor to create new lists when this list is modified.</p>
	 * <p>All subclasses of SList should use this method instead of a constructor.
	 * This method can be overridden in a subclass and made to return objects of the subclass's type.
	 * This avoids the new to re-write the modifying methods such as {@link #add} and {@link #remove} in every subclass.</p>
	 * 
	 * @param length the number of elements in the list
	 * @param first the first node in the list
	 */
	protected SList<O> makeList(int length, Node<O> first){
		return new SList<O>(length, first);
	}
	
	/**
	 * <p>Construct an empty static list.</p>
	 */
	public SList(){
		super();
	}
	
	/**
	 * <p>Construct a static list which initially contains a given set of elements.</p>
	 * <p>This constructor is more efficient in both time and space than creating an empty static list and then adding the elements to it.</p>
	 * 
	 * @param elements the elements the list will initially contain
	 */
	public SList(O[] elements){
		super(elements);
	}
	
	/**
	 * <p>Add an element to the static list in O(1) time.</p>
	 * <p>The first element (index 0) is added first, which means that calling {@link #getLast} on the returned list will return <code>elements[elements.length - 1]</code>.<p>
	 * 
	 * @param element the element to be added
	 * @return a clone of the list with <tt>element</tt> added to it
	 */
	public SList<O> add(O element){
		return makeList(length + 1, new Node<O>(element, last));
	}
	
	/**
	 * <p>Add x elements to the static list in O(x) time.</p>
	 * <p>This method is only marginally more efficient than multiple calls to {@link #add} and is provided for convenience only.</p>
	 * 
	 * @param elements an array of elements to be added
	 * @return a clone of the list with all the <tt>elements</tt> added to it
	 */
	public SList<O> add(O[] elements){
		Node<O> newFirst = last;
		for(int i=0; i<elements.length; i++)
			newFirst = new Node<O>(elements[i], newFirst);
		return makeList(length + elements.length, newFirst);
	}
	
	/**
	 * <p>Add every element from <tt>otherList</tt> to the static list.</p>
	 * <p>The first element added is the first element that was added to <tt>otherList</tt>, which means <code>result.{@link #getLast}==otherList.{@link #getLast}</code>.</p>
	 * 
	 * @param otherList a list of elements to be added to this list
	 * @return a clone of the list with each element from <tt>otherList</tt> added to it
	 */
	public SList<O> add(List<O> otherList){
		return makeList(length + otherList.length(), recursiveConcatenate(otherList.last));
	}
	
	/**
	 * <p>A recursive helper function for {@link #add(List)} that adds the first element of <tt>otherList</tt> first.</p>
	 * 
	 * @param node the current node in the list to be copied
	 * @return a new chain of nodes containing the elements of <tt>otherList</tt> in order and ending in {@link #first}.
	 */
	private final Node<O> recursiveConcatenate(Node<O> node){
		if(node.next == null)
			return new Node<O>(node.content, last);
		else
			return new Node<O>(node.content, recursiveConcatenate(node.next));
	}

	/**
	 * <p>Remove the first instance of a given object from the static list in O(n) time.  Objects are compared using <code>==</code>.</p>
	 * 
	 * @param element the object to remove
	 * @return a clone of the list which contains one less instance of <tt>element</tt>, or the same list if no instance of <tt>element</tt> was found
	 */
	public SList<O> remove(O element){
		Node<O> newFirst = removeFind(last, element);
		if(newFirst == last) return this;
		else return makeList(length - 1, newFirst);
	}
	
	/**
	 * <p>A recursive helper function for {@link #remove} that returns a chain of nodes with one less instance of <tt>object</tt>.</p>
	 * 
	 * @param node the current node in the chain
	 * @param object the object to be removed
	 * @return a copied chain of nodes (or the same chain if <tt>object</tt> was not found
	 */
	private final Node<O> removeFind(Node<O> node, O object){
		if(node == null)
			return null;
		else if(node.content == object)
			return node.next;
		else{
			Node<O> rest = removeFind(node.next, object);
			if(rest == node.next)
				return node;
			else
				return new Node<O>(node.content, rest);
		}
	}
	
	/**
	 * <p>Remove the first instance of a given object from the static list in O(n) time.  Objects are compared using {@link Object#equals}.</p>
	 * 
	 * @param element the object to remove
	 * @return a clone of the list which contains one less instance of <tt>element</tt>, or the same list if no instance of <tt>element</tt> was found
	 */
	public SList<O> removeEquals(O element){
		Node<O> newFirst = removeFindEquals(last, element);
		if(newFirst == last) return this;
		else return makeList(length - 1, newFirst);
	}
	
	/**
	 * <p>A recursive helper function for {@link #removeEquals} that returns a chain of nodes with one less instance of <tt>object</tt>.</p>
	 * 
	 * @param node the current node in the chain
	 * @param object the object to be removed
	 * @return a copied chain of nodes (or the same chain if <tt>object</tt> was not found
	 */
	private final Node<O> removeFindEquals(Node<O> node, O object){
		if(node == null)
			return null;
		else if(node.content.equals(object))
			return node.next;
		else{
			Node<O> rest = removeFindEquals(node.next, object);
			if(rest == node.next)
				return node;
			else
				return new Node<O>(node.content, rest);
		}
	}
	
	/**
	 * <p>Remove every instance of a given object from the static list in O(n) time.  Objects are compared using <code>==</code>.</p>
	 * <p>This method is more time efficient than multiple calls to {@link #remove}.</p>
	 * 
	 * @param element the object to remove every instance of
	 * @return a clone of the list which contains no instances of <tt>element</tt>, or the same list if no instances of <tt>element</tt> were found
	 */
	public SList<O> removeAll(O element){
		Counter lengthCounter = new Counter(length);
		Node<O> newFirst = removeFindAll(last, element, lengthCounter);
		if(newFirst == last) return this;
		else return makeList(lengthCounter.count, newFirst);
	}
	
	/**
	 * <p>An integer counter which can be passed like an object.</p>
	 * 
	 * @author Stephen G. Ware
	 */
	private final class Counter {
		public int count;
		public Counter(int count){
			this.count = count;
		}
	}
	
	/**
	 * <p>A recursive helper function for {@link #removeAll} that returns a chain of nodes with one less instance of <tt>object</tt>.</p>
	 * 
	 * @param node the current node in the chain
	 * @param object the object to be removed
	 * @param length a {@link Counter} for keeping track of the new length of the list
	 * @return a copied chain of nodes (or the same chain if <tt>object</tt> was not found
	 */
	private final Node<O> removeFindAll(Node<O> node, O object, Counter length){
		if(node == null)
			return null;
		else if(node.content == object){
			length.count--;
			return removeFindAll(node.next, object, length);
		}
		else{
			Node<O> rest = removeFindAll(node.next, object, length);
			if(rest == node.next)
				return node;
			else
				return new Node<O>(node.content, rest);
		}
	}
	
	/**
	 * <p>Remove every instance of a given object from the list in O(n) time.  Objects are compared using {@link Object#equals}.</p>
	 * <p>This method is more time efficient than multiple calls to {@link #removeEquals}.</p>
	 * 
	 * @param element the object to remove every instance of
	 * @return a clone of this list which contains no instances of <tt>element</tt>, or the same list if no instances of <tt>element</tt> were found
	 */
	public SList<O> removeAllEquals(O element){
		Counter lengthCounter = new Counter(length);
		Node<O> newFirst = removeFindAllEquals(last, element, lengthCounter);
		if(newFirst == last) return this;
		else return makeList(lengthCounter.count, newFirst);
	}
	
	/**
	 * <p>A recursive helper function for {@link #removeAllEquals} that returns a chain of nodes with one less instance of <tt>object</tt>.</p>
	 * 
	 * @param node the current node in the chain
	 * @param object the object to be removed
	 * @param length a {@link Counter} for keeping track of the new length of the list
	 * @return a copied chain of nodes (or the same chain if <tt>object</tt> was not found
	 */
	private final Node<O> removeFindAllEquals(Node<O> node, O object, Counter length){
		if(node == null)
			return null;
		else if(node.content.equals(object)){
			length.count--;
			return removeFindAllEquals(node.next, object, length);
		}
		else{
			Node<O> rest = removeFindAllEquals(node.next, object, length);
			if(rest == node.next)
				return node;
			else
				return new Node<O>(node.content, rest);
		}
	}
	
	/**
	 * <p>Return a deep copy of the dynamic list.</p>
	 * <p>The clone list can be modified without affecting the original.</p>
	 */
	public SList<O> clone(){
		return makeList(length, last);
	}
}
