package unit_tests;

import sdds.HashTable;
import java.util.Iterator;
import java.util.NoSuchElementException;

import junit.framework.TestCase;

/**
 * <p>Tests common to all subclasses of {@link sdds.HashTable}</p>
 * 
 * @author Stephen G. Ware
 */
public abstract class HashTableTest extends TestCase {

	protected static final Item[] keys = makeKeys();
	protected static final Item[] values = makeValues();
	protected static final Item[] moreValues = makeMoreValues();
	
	private static final Item[] makeKeys(){
		Item[] keys = new Item[100];
		for(int i=0; i<100; i++)
			keys[i] = new Item("k" + i);
		return keys;
	}
	
	private static final Item[] makeValues(){
		Item[] values = new Item[100];
		for(int i=0; i<100; i++)
			values[i] = new Item("v" + i);
		return values;
	}
	
	private static final Item[] makeMoreValues(){
		Item[] values = new Item[100];
		for(int i=0; i<100; i++)
			values[i] = new Item("vv" + i);
		return values;
	}
	
	public void testAddGet(HashTable<Item, Item> hashTable){
		System.out.println("\n=== After an element is added to a " + hashTable.getClass().getSimpleName() + " it can be gotten by its key. ===");
		System.out.print("Begin with an empty hash table.  It should have 0 mappings: ");
		assertEquals(hashTable.size(), 0);
		System.out.println(hashTable.size());
		System.out.print("Add " + values[1] + " keyed on " + keys[1] + ".  The table should now have 1 mapping: ");
		hashTable = hashTable.add(keys[1], values[1]);
		assertEquals(hashTable.size(), 1);
		System.out.println(hashTable.size());
		System.out.print("Getting key " + keys[1] + " should return the value " + values[1] + ": ");
		Item gotten = hashTable.get(keys[1]);
		assertSame(gotten, values[1]);
		System.out.println(gotten);
		System.out.print("Getting key " + keys[2] + " should return null: ");
		gotten = hashTable.get(keys[2]);
		assertNull(gotten);
		System.out.println(gotten);
	}
	
	public void test100AddGet(HashTable<Item, Item> hashTable){
		System.out.println("\n=== Many add operations should cause a " + hashTable.getClass().getSimpleName() + " to rehash. ===");
		System.out.println("Begin with an empty hash table.  For " + values[0] + " through " + values[99] + ", add items with an even ID to the table");
		System.out.println("Each item is keyed on a k item whose ID is the ID of the v item - 100.");
		for(int i=0; i<keys.length; i+=2){
			if(i % 2 == 0)
				hashTable = hashTable.add(keys[i], values[i]);
		}
		System.out.print("The table should have 50 mappings: ");
		assertEquals(hashTable.size(), 50);
		System.out.println(hashTable.size());
		System.out.print("Get every key from " + keys[0] + " through " + keys[99] + ".  Odd ID's should return null.  Even ID's should return a v item: ");
		Item gotten;
		for(int i=0; i<keys.length; i++){
			gotten = hashTable.get(keys[i]);
			if(i % 2 == 0)
				assertSame(gotten, values[i]);
			else
				assertNull(gotten);
		}
		System.out.println("they did.");
	}
	
	public void test100AddGetEquals(HashTable<Item, Item> hashTable){
		System.out.println("\n=== Many add operations using equals() should cause a " + hashTable.getClass().getSimpleName() + " to rehash. ===");
		System.out.println("Begin with an empty hash table.  For " + values[0] + " through " + values[99] + ", add items with an even ID to the table.");
		System.out.println("Each item is keyed on a k item whose ID is 100 + the ID of the v item.");
		for(int i=0; i<keys.length; i+=2){
			if(i % 2 == 0)
				hashTable = hashTable.addEquals(keys[i], values[i]);
		}
		System.out.print("The table should have 50 mappings: ");
		assertEquals(hashTable.size(), 50);
		System.out.println(hashTable.size());
		System.out.print("Get every key from " + keys[0] + " through " + keys[99] + " using equals().  Odd ID's should return null.  Even ID's should return a v item: ");
		Item gotten;
		for(int i=0; i<keys.length; i++){
			gotten = hashTable.getEquals(keys[i]);
			if(i % 2 == 0)
				assertSame(gotten, values[i]);
			else
				assertNull(gotten);
		}
		System.out.println("they did.");
	}
	
	public void testReplace(HashTable<Item, Item> hashTable){
		System.out.println("\n=== Adding a key or value that already exists in a " + hashTable.getClass().getSimpleName() + " should overwrite the original mapping. ===");
		System.out.print("Begin with an empty hash table.  It should have 0 mappings: ");
		assertEquals(hashTable.size(), 0);
		System.out.println(hashTable.size());
		System.out.print("Add " + values[1] + " keyed on " + keys[1] + " and " + values[2] + " keyed on " + keys[2] + ".  These keys sould have the same hash code: ");
		assertEquals(keys[1].hashCode(), keys[2].hashCode());
		hashTable = hashTable.add(keys[1], values[1]);
		hashTable = hashTable.add(keys[2], values[2]);
		System.out.println("they do");
		System.out.print("The table should now have 2 mappings: ");
		assertEquals(hashTable.size(), 2);
		System.out.println(hashTable.size());
		System.out.print("Add " + values[3] + " keyed on " + keys[1] + ".  The table should still have only 2 mappings: ");
		hashTable = hashTable.add(keys[1], values[3]);
		assertEquals(hashTable.size(), 2);
		System.out.println(hashTable.size());
		System.out.print("Getting key " + keys[1] + " should return " + values[3] + ": ");
		Item gotten = hashTable.get(keys[1]);
		assertSame(gotten, values[3]);
		System.out.println(gotten);
		Item newKey = new Item("k1");
		System.out.print("Add " + values[3] + " keyed on " + newKey + " using equals().  The table should still have only 2 mappings: ");
		hashTable = hashTable.addEquals(newKey, values[3]);
		System.out.println(hashTable.size());
		assertEquals(hashTable.size(), 2);
		System.out.print("Getting the original key " + keys[1] + " should return null: ");
		gotten = hashTable.get(keys[1]);
		System.out.println(gotten);
		assertNull(gotten);
		System.out.print("Getting the new key " + newKey + " should return " + values[3] + ": ");
		gotten = hashTable.get(newKey);
		assertSame(gotten, values[3]);
		System.out.println(gotten);
	}
	
	public void testMultiAdd(HashTable<Item, Item> hashTable){
		System.out.println("\n=== Many key / value pairs can be added into a " + hashTable.getClass().getSimpleName() + " at one time. ===");
		System.out.print("Begin with an empty hash table.  Add an array of 0 values keyed on 0 items.  The hash table should not have changed: ");
		HashTable<Item, Item> backup = hashTable;
		hashTable = hashTable.add(new Item[0], new Item[0]);
		assertSame(backup, hashTable);
		System.out.println("it did not");
		System.out.print("Add an array of 0 values keyed on 0 items using equals().  The hash table should not have changed: ");
		backup = hashTable;
		hashTable = hashTable.addEquals(new Item[0], new Item[0]);
		assertSame(backup, hashTable);
		System.out.println("it did not");
		System.out.println("Add an array of 100 v items keyed on 100 k items.");
		hashTable = hashTable.add(keys, values);
		System.out.print("The hash table should have 100 mappings: ");
		System.out.println(hashTable.size());
		assertEquals(hashTable.size(), 100);
		System.out.print("Getting each k item should return its associated v item: ");
		for(int i=0; i<keys.length; i++)
			assertEquals(hashTable.get(keys[i]), values[i]);
		System.out.println("it does");
	}
	
	public void test100MultiReplace(HashTable<Item, Item> hashTable){
		System.out.println("\n=== Many key / value pairs can be added into a " + hashTable.getClass().getSimpleName() + " at one time, replacing values as nescessary. ===");
		System.out.println("Begin with an empty hash table.  Add an array of 100 v items keyed on 100 k items.");
		hashTable = hashTable.add(keys, values);
		System.out.print("The hash table should have 100 mappings: ");
		System.out.println(hashTable.size());
		assertEquals(hashTable.size(), 100);		
		System.out.print("Add an array of 100 new v items keyed on the same k items.  The hash table should still have 100 mappings: ");
		hashTable = hashTable.add(keys, moreValues);		
		System.out.println(hashTable.size());
		assertEquals(hashTable.size(), 100);
	}
	
	public void testMultiAddEquals(HashTable<Item, Item> hashTable){
		System.out.println("\n=== Many key / value pairs can be added into a " + hashTable.getClass().getSimpleName() + " at one time using equals(). ===");
		System.out.println("Begin with an empty hash table.  Add " + values[0] + " keyed on " + keys[0] + " using equals().");
		hashTable = hashTable.addEquals(keys[0], values[0]);
		System.out.println("Add an array of 100 v items (including " + values[0] + ") keyed on 100 k items (including " + keys[0] + ") using equals().");
		hashTable = hashTable.addEquals(keys, values);
		System.out.print("The hash table should have 100 mappings: ");
		System.out.println(hashTable.size());
		assertEquals(hashTable.size(), 100);
		System.out.print("Getting each k item should return its associated v item: ");
		for(int i=0; i<keys.length; i++)
			assertEquals(hashTable.get(keys[i]), values[i]);
		System.out.println("it does");
	}
	
	public void testMultiAddDifferentSizeArrays(HashTable<Item, Item> hashTable){
		System.out.println("\n=== Adding multiple key / value pairs into a " + hashTable.getClass().getSimpleName() + " should fail if the number of keys is not equal to the number of values. ===");
		System.out.println("Begin with an empty hash table.");
		System.out.print("Add 100 v items keyed on only 99 k items: ");
		Item[] keys = new Item[99];
		boolean thrown = false;
		try{
			hashTable = hashTable.add(keys, values);
		}
		catch(ArrayIndexOutOfBoundsException ex){
			System.out.println(ex.getMessage());
			thrown = true;
		}
		assertTrue(thrown);
		System.out.print("Add 100 v items keyed on only 99 k items using equals(): ");
		try{
			hashTable = hashTable.addEquals(keys, values);
		}
		catch(ArrayIndexOutOfBoundsException ex){
			System.out.println(ex.getMessage());
			return;
		}
	}
	
	public void testRemove(HashTable<Item, Item> hashTable){
		System.out.println("\n=== Key mappings can be removed from a " + hashTable.getClass().getSimpleName() + ". ===");
		System.out.print("Begin with an empty hash table.  Removing the key " + keys[1] + " should result in no change: ");
		assertSame(hashTable, hashTable.remove(keys[1]));
		System.out.println("no change\nAdd item " + values[1] + " keyed on " + keys[1] + ", item " + values[2] + " keyed on " + keys[2] + ", and " + values[3] + " keyed on " + keys[3] + ".");
		hashTable = hashTable.add(keys[1], values[1]);
		hashTable = hashTable.add(keys[2], values[2]);
		hashTable = hashTable.add(keys[3], values[3]);
		System.out.println("3 mappings should exist: " + hashTable.size());
		assertEquals(hashTable.size(), 3);
		Item gotten;
		System.out.print("Getting " + keys[1] + " should return " + values[1] + ": ");
		gotten = hashTable.get(keys[1]);
		System.out.println(gotten);
		assertSame(gotten, values[1]);
		System.out.print("Getting " + keys[3] + " should return " + values[3] + ": ");
		gotten = hashTable.get(keys[3]);
		System.out.println(gotten);
		assertSame(gotten, values[3]);
		System.out.print("Remove key " + keys[1] + ".  There should now be 2 mappings: ");
		hashTable = hashTable.remove(keys[1]);
		System.out.println(hashTable.size());
		assertEquals(hashTable.size(), 2);
		System.out.print("Getting " + keys[1] + " should return null: ");
		gotten = hashTable.get(keys[1]);
		assertNull(gotten);
		System.out.println("null");
		System.out.print("Removing key " + keys[1] + " should result in no change: ");
		assertSame(hashTable, hashTable.remove(keys[1]));
		System.out.println("no change");
		System.out.print("Remove key " + keys[3] + ".  There should now be 1 mapping: ");
		hashTable = hashTable.remove(keys[3]);
		System.out.println(hashTable.size());
		assertEquals(hashTable.size(), 1);
		System.out.print("Getting " + keys[3] + " should return null: ");
		gotten = hashTable.get(keys[3]);
		assertNull(gotten);
		System.out.println("null");
	}
	
	public void testMultiRemove(HashTable<Item, Item> hashTable){
		System.out.println("\n=== Multiple key mappings can be removed from a " + hashTable.getClass().getSimpleName() + ". ===");
		System.out.print("Begin with an empty hash table.");
		Item[] threeKeys = new Item[3];
		threeKeys[0] = keys[1];
		threeKeys[1] = keys[2];
		threeKeys[2] = keys[3];
		System.out.print("  Remove all of these keys at the same time:");
		for(Item key : threeKeys)
			System.out.print(" " + key);
		System.out.print(": ");
		assertSame(hashTable, hashTable.remove(threeKeys));
		System.out.println("no change\nAdd item " + values[1] + " keyed on " + keys[1] + " and item " + values[2] + " keyed on " + keys[2] + ".");
		hashTable = hashTable.add(keys[1], values[1]);
		hashTable = hashTable.add(keys[2], values[2]);
		System.out.println("The hash table should have 2 mappings: " + hashTable.size());
		assertEquals(hashTable.size(), 2);
		System.out.print("Remove all of these keys at the same time:");
		for(Item key : threeKeys)
			System.out.print(" " + key);
		System.out.println();
		hashTable = hashTable.remove(threeKeys);
		System.out.println("The hash table should now have 0 mappings: " + hashTable.size());
		assertEquals(hashTable.size(), 0);
	}
	
	public void testRemoveEquals(HashTable<Item, Item> hashTable){
		System.out.println("\n=== Key mappings can be removed from a " + hashTable.getClass().getSimpleName() + " using equals(). ===");
		System.out.print("Begin with an empty hash table.  Removing the key " + keys[1] + " using equals() should result in no change: ");
		assertSame(hashTable, hashTable.removeEquals(keys[1]));
		System.out.println("no change\nAdd item " + values[1] + " keyed on " + keys[1] + ", item " + values[2] + " keyed on " + keys[2] + ", and " + values[3] + " keyed on " + keys[3] + ".");
		hashTable = hashTable.add(keys[1], values[1]);
		hashTable = hashTable.add(keys[2], values[2]);
		hashTable = hashTable.add(keys[3], values[3]);
		System.out.println("3 mappings should exist: " + hashTable.size());
		assertEquals(hashTable.size(), 3);
		Item gotten;
		Item newKey1 = new Item("k1");
		Item newKey3 = new Item("k3");
		System.out.print("Getting " + newKey1 + " using equals() should return " + values[1] + ": ");
		gotten = hashTable.getEquals(newKey1);
		System.out.println(gotten);
		assertSame(gotten, values[1]);
		System.out.print("Getting " + newKey3 + " using equals() should return " + values[3] + ": ");
		gotten = hashTable.getEquals(newKey3);
		System.out.println(gotten);
		assertSame(gotten, values[3]);
		System.out.print("Remove key " + newKey1 + " using equals().  There should now 2 mappings: ");
		hashTable = hashTable.removeEquals(newKey1);
		System.out.println(hashTable.size());
		assertEquals(hashTable.size(), 2);
		System.out.print("Getting " + newKey1 + " using equals() should return null: ");
		gotten = hashTable.getEquals(newKey1);
		assertNull(gotten);
		System.out.println("null");
		System.out.print("Removing " + newKey1 + " using equals() should result in no change: ");
		assertSame(hashTable, hashTable.removeEquals(newKey1));
		System.out.println("no change");
		System.out.print("Remove key " + newKey3 + " using equals().  There should now be 1 mapping: ");
		hashTable = hashTable.removeEquals(newKey3);
		System.out.println(hashTable.size());
		assertEquals(hashTable.size(), 1);
		System.out.print("Getting " + newKey3 + " using equals() should return null: ");
		gotten = hashTable.getEquals(newKey3);
		assertNull(gotten);
		System.out.println("null");
	}
	
	public void testMultiRemoveEquals(HashTable<Item, Item> hashTable){
		System.out.println("\n=== Multiple key mappings can be removed from a " + hashTable.getClass().getSimpleName() + " using equals(). ===");
		System.out.print("Begin with an empty hash table.");
		Item[] threeKeys = new Item[3];
		threeKeys[0] = new Item("k1");
		threeKeys[1] = new Item("k2");
		threeKeys[2] = new Item("k3");
		System.out.print("Remove all of these keys at the same time using equals():");
		for(Item key : threeKeys)
			System.out.print(" " + key);
		System.out.println();
		assertSame(hashTable, hashTable.removeEquals(threeKeys));
		System.out.println("no change\nAdd item " + values[1] + " keyed on " + keys[1] + " and item " + values[2] + " keyed on " + keys[2] + ".");
		hashTable = hashTable.add(keys[1], values[1]);
		hashTable = hashTable.add(keys[2], values[2]);
		System.out.println("2 mappings should exist: " + hashTable.size());
		assertEquals(hashTable.size(), 2);
		System.out.print("Remove all of these keys at the same time using equals():");
		for(Item key : threeKeys)
			System.out.print(" " + key);
		System.out.println();
		hashTable = hashTable.removeEquals(threeKeys);
		System.out.println("0 mappings should remain: " + hashTable.size());
		assertEquals(hashTable.size(), 0);
	}
	
	public void testKeyIterator(HashTable<Item, Item> hashTable){
		System.out.println("\n=== A " + hashTable.getClass().getSimpleName() + "'s key iterator should yield all of its keys. ===");
		System.out.print("Begin with an empty hash table.  Its key iterator should return no values: ");
		Iterator<Item> iterator = hashTable.keys();
		assertFalse(iterator.hasNext());
		System.out.println("no values");
		System.out.print("The iterator should throw an exception if we try to get any values: ");
		try{
			iterator.next();
			fail();
		}
		catch(NoSuchElementException ex){
			System.out.println(ex.getMessage());
		}
		System.out.println("For " + values[0] + " through " + values[99] + ", add items with an even ID to the table keyed on a k item whose ID equals the ID of the v item - 100.");
		for(int i=0; i<100; i+=2)
			hashTable = hashTable.add(keys[i], values[i]);
		System.out.print("The table should have 50 mappings: ");
		assertEquals(hashTable.size(), 50);
		System.out.println(hashTable.size());
		System.out.print("Get an iterator for all the keys in the table.  The iterator should throw an exception if we try to remove a value: ");
		try{
			iterator.remove();
			fail();
		}
		catch(UnsupportedOperationException ex){
			System.out.println(ex.getMessage());
		}
		System.out.print("Check that each k item returns its associated v item: ");
		iterator = hashTable.keys();
		Item key;
		int count = 0;
		while(iterator.hasNext()){
			key = iterator.next();
			assertEquals(key.id + 100, hashTable.get(key).id);
			count++;
		}
		System.out.println("they do.");
		System.out.print("The iterator should throw an exception if we try to get any more values: ");
		try{
			iterator.next();
			fail();
		}
		catch(java.util.NoSuchElementException ex){
			System.out.println(ex.getMessage());
		}
		System.out.print("The iterator should have yielded 50 keys: ");
		assertEquals(count, 50);
		System.out.println(count);
	}
	
	public void testValueIterator(HashTable<Item, Item> hashTable){
		System.out.println("\n=== A " + hashTable.getClass().getSimpleName() + "'s value iterator should yield all of its values. ===");
		System.out.print("Begin with an empty hash table.  Its value iterator should return no values: ");
		Iterator<Item> iterator = hashTable.values();
		assertFalse(iterator.hasNext());
		System.out.println("no values");
		System.out.print("The iterator should throw an exception if we try to get any values: ");
		try{
			iterator.next();
			fail();
		}
		catch(NoSuchElementException ex){
			System.out.println(ex.getMessage());
		}
		System.out.println("For " + values[0] + " through " + values[99] + ", add items with an even ID to the table keyed on a k item whose ID is the ID of the v item - 100.");
		for(int i=0; i<100; i+=2)
			hashTable = hashTable.add(keys[i], values[i]);
		System.out.print("The table should have 50 mappings: ");
		assertEquals(hashTable.size(), 50);
		System.out.println(hashTable.size());
		System.out.print("Get an iterator for all the values in the hash table.  The iterator should throw an exception if we try to remove a value: ");
		iterator = hashTable.values();
		try{
			iterator.remove();
			fail();
		}
		catch(UnsupportedOperationException ex){
			System.out.println(ex.getMessage());
		}
		System.out.print("Check that each value has an even ID: ");
		int count = 0;
		while(iterator.hasNext()){
			assertTrue(iterator.next().id % 2 == 0);
			count++;
		}
		System.out.println("they do.");
		System.out.print("The iterator should throw an exception if we try to get any more values: ");
		try{
			iterator.next();
			fail();
		}
		catch(java.util.NoSuchElementException ex){
			System.out.println(ex.getMessage());
		}
		System.out.print("The iterator should have yielded 50 keys: ");
		assertEquals(count, 50);
		System.out.println(count);
	}
	
	public void testEquals(HashTable<Item, Item> hashTable1, HashTable<Item, Item> hashTable2){
		System.out.println("\n=== " + hashTable1.getClass().getSimpleName() + "s can be compared using equals(). ===");
		assertNotSame(hashTable1, hashTable2);
		System.out.print("Begin with two empty hash tables.  The first should not be equal to null: ");
		assertFalse(hashTable1.equals(null));
		System.out.println("it is not\nAdd " + values[1] + " keyed on " + keys[1] + " and " + values[2] + " keyed on " + keys[2] + " to the first hash table.");
		hashTable1 = hashTable1.add(keys[1], values[1]);
		hashTable1 = hashTable1.add(keys[2], values[2]);
		System.out.println("Add " + values[2] + " keyed on " + keys[2] + " to the second hash table.");
		hashTable2 = hashTable2.add(keys[2], values[2]);
		System.out.print("The hash tables should not be equal: ");
		assertFalse(hashTable1.equals(hashTable2));
		System.out.println("they are not\nAdd " + values[3] + " keyed on " + keys[1] + " to the second hash table.");
		hashTable2 = hashTable2.add(keys[1], values[3]);
		System.out.print("The hash tables should not be equal: ");
		assertFalse(hashTable1.equals(hashTable2));
		System.out.println("they are not\nAdd " + values[1] + " keyed on " + keys[1] + " to the second hash table.");
		hashTable2 = hashTable2.add(keys[1], values[1]);
		System.out.print("The hash tables should be equal: ");
		assertTrue(hashTable1.equals(hashTable2));
		System.out.println("they are");
	}
	
	public void testClone(HashTable<Item, Item> hashTable){
		System.out.println("\n=== " + hashTable.getClass().getSimpleName() + "s can be cloned. ===");
		System.out.println("Begin with an empty hash table.  Add " + values[1] + " keyed on " + keys[1] + " and " + values[2] + " keyed on " + keys[2] + " to the first hash table.");
		hashTable = hashTable.add(keys[1], values[1]);
		hashTable = hashTable.add(keys[2], values[2]);
		System.out.print("Clone the hash table.  The original and the clone should not be the same object: ");
		HashTable<Item, Item> clone = hashTable.clone();
		assertNotSame(hashTable, clone);
		System.out.print("they are not\nThe original and the clone should be equal: ");
		assertEquals(hashTable, clone);
		System.out.print("they are\nRemove key " + keys[1] + " from the clone.  The original should be unchanged: ");
		clone = clone.remove(keys[1]);
		assertEquals(clone.size(), 1);
		assertEquals(hashTable.size(), 2);
		System.out.println("no change");
	}
	
	public void testToString(HashTable<Item, Item> hashTable){
		System.out.println("\n=== Test toString() for " + hashTable.getClass().getSimpleName() + ". ===");
		System.out.println("Begin with an empty hash table.  Add " + values[0] + " through " + values[4] + " keyed on " + keys[0] + " through " + keys[4]);
		for(int i=0; i<5; i++)
			hashTable = hashTable.add(keys[i], values[i]);
		System.out.println("Test the toString() method:\n");
		System.out.println(hashTable.toString());
	}
}
